# -*- coding: utf-8 -*-
#
#   STUN: Session Traversal Utilities for NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Session Traversal Utilities for NAT
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    [RFC] https://www.ietf.org/rfc/rfc5389.txt
    [RFC] https://www.ietf.org/rfc/rfc3489.txt
"""

from typing import Optional

from udp.tlv.data import bytes_to_int, uint8_to_bytes, uint16_to_bytes, uint32_to_bytes
from udp.tlv import UInt16Data, UInt32Data
from udp.tlv import Tag, Length, Value, TagLengthValue

"""
    STUN Attributes
    ~~~~~~~~~~~~~~~

   After the STUN header are zero or more attributes.  Each attribute
   MUST be TLV encoded, with a 16-bit type, 16-bit length, and value.
   Each STUN attribute MUST end on a 32-bit boundary.  As mentioned
   above, all fields in an attribute are transmitted most significant
   bit first.

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |         Type                  |            Length             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Value (variable)                ....
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

                    Figure 4: Format of STUN Attributes
"""


class AttributeValue(Value):

    @classmethod
    def parse(cls, data: bytes, tag: Tag, length: Length=None):
        if cls is AttributeValue:
            # get attribute parser with type
            assert isinstance(tag, AttributeType), 'attribute type error: %s' % tag
            parser = s_attribute_parsers.get(tag)
            if parser is not None:
                # create instance by subclass
                return parser.parse(data=data, tag=tag, length=length)
        # check length
        if length is None or length.value == 0:
            return None
        else:
            length = length.value
            data_len = len(data)
            if data_len < length:
                return None
            elif data_len > length:
                data = data[:length]
        return cls(data=data)


class AttributeLength(UInt16Data, Length):

    def __init__(self, value: int, data: bytes=None):
        if data is None:
            data = uint16_to_bytes(value)
        super().__init__(data=data, value=value)

    @classmethod
    def parse(cls, data: bytes, tag: Tag):
        length = super().parse(data=data, tag=tag)
        if length is not None:
            assert length.value & 0x0003 == 0, 'attribute length error: %d' % length.value
            return length


class AttributeType(UInt16Data, Tag):

    def __init__(self, value: int, data: bytes=None, name: str=None):
        if data is None:
            data = uint16_to_bytes(value)
        super().__init__(value=value, data=data)
        self.__name = name
        s_attribute_types[value] = self

    def __str__(self):
        # clazz = self.__class__.__name__
        if self.__name is None:
            return '"AttributeType-0x%04X"' % self.value
        else:
            return '"%s"' % self.__name

    def __repr__(self):
        # clazz = self.__class__.__name__
        if self.__name is None:
            return '"AttributeType-0x%04X"' % self.value
        else:
            return '"%s"' % self.__name

    def __hash__(self) -> int:
        return hash(self.value)

    @classmethod
    def parse(cls, data: bytes):
        data_len = len(data)
        if data_len < 2:
            return None
        elif data_len > 2:
            data = data[:2]
        value = bytes_to_int(data=data)
        t = s_attribute_types.get(value)
        if t is None:
            return cls(value=value, data=data)
        else:
            return t


# Attribute Types in STUN message
s_attribute_types = {}

# Comprehension-required range (0x0000-0x7FFF)
# Comprehension-optional range (0x8000-0xFFFF)

# [RFC-3489]
MappedAddress = AttributeType(0x0001, name='MAPPED-ADDRESS')
ResponseAddress = AttributeType(0x0002, name='RESPONSE-ADDRESS')
ChangeRequest = AttributeType(0x0003, name='CHANGE-REQUEST')
SourceAddress = AttributeType(0x0004, name='SOURCE-ADDRESS')
ChangedAddress = AttributeType(0x0005, name='CHANGED-ADDRESS')
Username = AttributeType(0x0006, name='USERNAME')
Password = AttributeType(0x0007, name='PASSWORD')
MessageIntegrity = AttributeType(0x0008, name='MESSAGE-INTEGRITY')
ErrorCode = AttributeType(0x0009, name='ERROR-CODE')
UnknownAttributes = AttributeType(0x000A, name='UNKNOWN-ATTRIBUTES')
ReflectedFrom = AttributeType(0x000B, name='REFLECTED-FROM')

# [RFC-5389]
Realm = AttributeType(0x0014, name='REALM')
Nonce = AttributeType(0x0015, name='NONCE')
XorMappedAddress = AttributeType(0x0020, name='XOR-MAPPED-ADDRESS(0020)')

XorMappedAddress2 = AttributeType(0x8020, name='XOR-MAPPED-ADDRESS(8020)')
XorOnly = AttributeType(0x8021, name='XOR-ONLY')
Software = AttributeType(0x8022, name='SOFTWARE')
AlternateServer = AttributeType(0x8023, name='ALTERNATE-SERVER')
Fingerprint = AttributeType(0x8028, name='FINGERPRINT')


class Attribute(TagLengthValue):

    def __init__(self, tag: AttributeType, value: AttributeValue, data: bytes=None):
        if data is None:
            data = tag.data + uint16_to_bytes(len(value.data)) + value.data
        super().__init__(data=data, tag=tag, value=value)

    def __str__(self):
        clazz = self.__class__.__name__
        return '/* %s */ %s: %s' % (clazz, self.tag, self.value)

    def __repr__(self):
        clazz = self.__class__.__name__
        return '/* %s */ %s: %s' % (clazz, self.tag, self.value)

    @classmethod
    def parse_tag(cls, data: bytes) -> Optional[AttributeType]:
        return AttributeType.parse(data=data)

    @classmethod
    def parse_length(cls, data: bytes, tag: AttributeType) -> Optional[AttributeLength]:
        return AttributeLength.parse(data=data, tag=tag)

    @classmethod
    def parse_value(cls, data: bytes, tag: AttributeType, length: AttributeLength = None) -> Optional[Value]:
        return AttributeValue.parse(data=data, tag=tag, length=length)


"""
Rosenberg, et al.           Standards Track                    [Page 32]

RFC 5389                          STUN                      October 2008


   The format of the MAPPED-ADDRESS attribute is:

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |0 0 0 0 0 0 0 0|    Family     |           Port                |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |                 Address (32 bits or 128 bits)                 |
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

               Figure 5: Format of MAPPED-ADDRESS Attribute

   The address family can take on the following values:

   0x01:IPv4
   0x02:IPv6

   The first 8 bits of the MAPPED-ADDRESS MUST be set to 0 and MUST be
   ignored by receivers.  These bits are present for aligning parameters
   on natural 32-bit boundaries.

   This attribute is used only by servers for achieving backwards
   compatibility with RFC 3489 [RFC3489] clients.
"""


class MappedAddressValue(AttributeValue):
    """
    15.1.  MAPPED-ADDRESS

        The MAPPED-ADDRESS attribute indicates a reflexive transport address
        of the client.  It consists of an 8-bit address family and a 16-bit
        port, followed by a fixed-length value representing the IP address.
        If the address family is IPv4, the address MUST be 32 bits.  If the
        address family is IPv6, the address MUST be 128 bits.  All fields
        must be in network byte order.
    """
    family_ipv4 = 0x01
    family_ipv6 = 0x02

    def __init__(self, data: bytes, ip: str, port: int, family: int):
        super().__init__(data=data)
        self.__family = family
        self.__port = port
        self.__ip = ip

    def __str__(self):
        return '"%s:%d"' % (self.ip, self.port)

    def __repr__(self):
        return '"%s:%d"' % (self.ip, self.port)

    @property
    def family(self) -> int:
        return self.__family

    @property
    def port(self) -> int:
        return self.__port

    @property
    def ip(self) -> str:
        return self.__ip

    @classmethod
    def ip_to_bytes(cls, ip: str, family: int) -> bytes:
        if family == cls.family_ipv4:
            # IPv4
            array = ip.split('.')
            assert len(array) == 4, 'IP address error: %s' % ip
            return bytes([int(x) for x in array])
        elif family == cls.family_ipv6:
            # TODO: IPv6
            assert False, 'IPv6 not support yet'
        else:
            raise ValueError('unknown address family: %d' % family)

    @classmethod
    def bytes_to_ip(cls, address: bytes, family: int) -> str:
        # check address family
        if family == cls.family_ipv4:
            assert len(address) == 4, 'IPv4 data error: %s' % address
            # IPv4
            return '.'.join([
                str(bytes_to_int(address[0:1])),
                str(bytes_to_int(address[1:2])),
                str(bytes_to_int(address[2:3])),
                str(bytes_to_int(address[3:4])),
            ])
        elif family == cls.family_ipv6:
            assert len(address) == 16, 'IPv6 data error: %s' % address
            # TODO: IPv6
            assert False, 'IPv6 not support yet'
        else:
            raise ValueError('unknown address family: %d' % family)

    @classmethod
    def parse(cls, data: bytes, tag: Tag, length: Length=None):
        # check length
        if length is None or (length.value != 8 and length.value != 20):
            # raise ValueError('mapped-address length error: %s' % length)
            return None
        else:
            length = length.value
            data_len = len(data)
            if data_len < length:
                # raise ValueError('data length error: %d, %d' % (data_len, length.value))
                return None
            elif data_len > length:
                data = data[:length]
        # checking
        if data[0] != 0:
            return None
        family = bytes_to_int(data[1:2])
        port = bytes_to_int(data[2:4])
        ip = cls.bytes_to_ip(address=data[4:], family=family)
        return cls(data=data, ip=ip, port=port, family=family)

    @classmethod
    def new(cls, ip: str, port: int, family: int=None):
        if family is None:
            family = cls.family_ipv4
        address = cls.ip_to_bytes(ip=ip, family=family)
        data = b'\0' + uint8_to_bytes(family) + uint16_to_bytes(port) + address
        return cls(data=data, ip=ip, port=port, family=family)


"""
   The format of the XOR-MAPPED-ADDRESS is:

      0                   1                   2                   3
      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |x x x x x x x x|    Family     |         X-Port                |
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     |                X-Address (Variable)
     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

             Figure 6: Format of XOR-MAPPED-ADDRESS Attribute

   The Family represents the IP address family, and is encoded
   identically to the Family in MAPPED-ADDRESS.


Rosenberg, et al.           Standards Track                    [Page 33]

RFC 5389                          STUN                      October 2008
"""


class XorMappedAddressValue(MappedAddressValue):
    """
    15.2.  XOR-MAPPED-ADDRESS

        The XOR-MAPPED-ADDRESS attribute is identical to the MAPPED-ADDRESS
        attribute, except that the reflexive transport address is obfuscated
        through the XOR function.

        X-Port is computed by taking the mapped port in host byte order,
        XOR'ing it with the most significant 16 bits of the magic cookie, and
        then the converting the result to network byte order.  If the IP
        address family is IPv4, X-Address is computed by taking the mapped IP
        address in host byte order, XOR'ing it with the magic cookie, and
        converting the result to network byte order.  If the IP address
        family is IPv6, X-Address is computed by taking the mapped IP address
        in host byte order, XOR'ing it with the concatenation of the magic
        cookie and the 96-bit transaction ID, and converting the result to
        network byte order.

        The rules for encoding and processing the first 8 bits of the
        attribute's value, the rules for handling multiple occurrences of the
        attribute, and the rules for processing address families are the same
        as for MAPPED-ADDRESS.

        Note: XOR-MAPPED-ADDRESS and MAPPED-ADDRESS differ only in their
        encoding of the transport address.  The former encodes the transport
        address by exclusive-or'ing it with the magic cookie.  The latter
        encodes it directly in binary.  RFC 3489 originally specified only
        MAPPED-ADDRESS.  However, deployment experience found that some NATs
        rewrite the 32-bit binary payloads containing the NAT's public IP
        address, such as STUN's MAPPED-ADDRESS attribute, in the well-meaning
        but misguided attempt at providing a generic ALG function.  Such
        behavior interferes with the operation of STUN and also causes
        failure of STUN's message-integrity checking.
    """

    @classmethod
    def xor(cls, data: bytes, factor: bytes) -> Optional[bytes]:
        if data[0] != 0:
            return None
        assert len(data) == 8 or len(data) == 20, 'address error: %s' % data
        assert len(factor) == 16, 'factor should be the "magic code" + "(96-bits) transaction ID"'
        array = bytearray(data)
        # X-Port
        array[2] ^= factor[1]
        array[3] ^= factor[0]
        # X-Address
        a_pos = len(array) - 1
        f_pos = 0
        while a_pos >= 4:
            array[a_pos] ^= factor[f_pos]
            a_pos -= 1
            f_pos += 1
        return bytes(array)

    @classmethod
    def new(cls, ip: str, port: int, family: int=None, factor: bytes=None):
        if family is None:
            family = cls.family_ipv4
        address = cls.ip_to_bytes(ip=ip, family=family)
        data = b'\0' + uint8_to_bytes(family) + uint16_to_bytes(port) + address
        assert factor is not None, 'missing factor'
        data = cls.xor(data=data, factor=factor)
        return cls(data=data, ip=ip, port=port, family=family)


class XorMappedAddressValue2(MappedAddressValue):
    """ https://tools.ietf.org/id/draft-ietf-behave-rfc3489bis-02.txt

    10.2.12  XOR-MAPPED-ADDRESS

        The XOR-MAPPED-ADDRESS attribute is only present in Binding
        Responses.  It provides the same information that is present in the
        MAPPED-ADDRESS attribute.  However, the information is encoded by

        performing an exclusive or (XOR) operation between the mapped address
        and the transaction ID.  Unfortunately, some NAT devices have been
        found to rewrite binary encoded IP addresses and ports that are
        present in protocol payloads.  This behavior interferes with the
        operation of STUN.  By providing the mapped address in an obfuscated
        form, STUN can continue to operate through these devices.

        The format of the XOR-MAPPED-ADDRESS is:

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |x x x x x x x x|    Family     |         X-Port                |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                X-Address (Variable)
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

        The Family represents the IP address family, and is encoded
        identically to the Family in MAPPED-ADDRESS.

        X-Port is equal to the port in MAPPED-ADDRESS, exclusive or'ed with
        most significant 16 bits of the transaction ID.  If the IP address
        family is IPv4, X-Address is equal to the IP address in MAPPED-
        ADDRESS, exclusive or'ed with the most significant 32 bits of the
        transaction ID.  If the IP address family is IPv6, the X-Address is
        equal to the IP address in MAPPED-ADDRESS, exclusive or'ed with the
        entire 128 bit transaction ID.
    """

    @classmethod
    def xor(cls, data: bytes, factor: bytes) -> Optional[bytes]:
        if data[0] != 0:
            return None
        assert len(data) == 8 or len(data) == 20, 'address error: %s' % data
        assert len(factor) == 16, 'factor should be the "magic code" + "(96-bits) transaction ID"'
        array = bytearray(data)
        # X-Port
        array[2] ^= factor[0]
        array[3] ^= factor[1]
        # X-Address
        length = len(array)
        a_pos = 4
        f_pos = 0
        while a_pos < length:
            array[a_pos] ^= factor[f_pos]
            a_pos += 1
            f_pos += 1
        return bytes(array)

    @classmethod
    def new(cls, ip: str, port: int, family: int=None, factor: bytes=None):
        if family is None:
            family = cls.family_ipv4
        address = cls.ip_to_bytes(ip=ip, family=family)
        data = b'\0' + uint8_to_bytes(family) + uint16_to_bytes(port) + address
        assert factor is not None, 'missing factor'
        data = cls.xor(data=data, factor=factor)
        return cls(data=data, ip=ip, port=port, family=family)


class ResponseAddressValue(MappedAddressValue):
    """
    11.2.2 RESPONSE-ADDRESS

        The RESPONSE-ADDRESS attribute indicates where the response to a
        Binding Request should be sent.  Its syntax is identical to MAPPED-
        ADDRESS.

    (Defined in RFC-3489, removed from RFC-5389)
    """
    pass


class ChangedAddressValue(MappedAddressValue):
    """
    11.2.3  CHANGED-ADDRESS

        The CHANGED-ADDRESS attribute indicates the IP address and port where
        responses would have been sent from if the "change IP" and "change
        port" flags had been set in the CHANGE-REQUEST attribute of the
        Binding Request.  The attribute is always present in a Binding
        Response, independent of the value of the flags.  Its syntax is
        identical to MAPPED-ADDRESS.

    (Defined in RFC-3489, removed from RFC-5389)
    """
    pass


class ChangeRequestValue(UInt32Data, AttributeValue):
    """
    11.2.4 CHANGE-REQUEST

       The CHANGE-REQUEST attribute is used by the client to request that
       the server use a different address and/or port when sending the
       response.  The attribute is 32 bits long, although only two bits (A
       and B) are used:

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 A B 0|
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

       The meaning of the flags is:

       A: This is the "change IP" flag.  If true, it requests the server
          to send the Binding Response with a different IP address than the
          one the Binding Request was received on.

       B: This is the "change port" flag.  If true, it requests the
          server to send the Binding Response with a different port than the
          one the Binding Request was received on.

    (Defined in RFC-3489, removed from RFC-5389)
    """

    def __init__(self, value: int, data: bytes=None):
        if data is None:
            data = uint32_to_bytes(value)
        super().__init__(data=data, value=value)

    def __str__(self):
        return '%d' % self.value

    def __repr__(self):
        return '%d' % self.value

    @classmethod
    def parse(cls, data: bytes, tag: Tag, length: Length=None):
        # check length
        if length is None or length.value != 4:
            # raise ValueError('Change-Request value error: %s' % length)
            return None
        else:
            length = length.value
            data_len = len(data)
            if data_len < length:
                # raise ValueError('data length error: %d, %d' % (data_len, length.value))
                return None
            elif data_len > length:
                data = data[:length]
        # get value
        value = bytes_to_int(data)
        if value == ChangeIPAndPort.value:
            return ChangeIPAndPort
        elif value == ChangeIP.value:
            return ChangeIP
        elif value == ChangePort.value:
            return ChangePort
        # else:
        #     # other values
        #     return ChangeRequestValue(value=value)


ChangeIP = ChangeRequestValue(0x00000004)
ChangePort = ChangeRequestValue(0x00000002)
ChangeIPAndPort = ChangeRequestValue(0x00000006)


class SourceAddressValue(MappedAddressValue):
    """
    11.2.5 SOURCE-ADDRESS

        The SOURCE-ADDRESS attribute is present in Binding Responses.  It
        indicates the source IP address and port that the server is sending
        the response from.  Its syntax is identical to that of MAPPED-
        ADDRESS.

    (Defined in RFC-3489, removed from RFC-5389)
    """
    pass


class SoftwareValue(AttributeValue):
    """
    15.10.  SOFTWARE

        The SOFTWARE attribute contains a textual description of the software
        being used by the agent sending the message.  It is used by clients
        and servers.  Its value SHOULD include manufacturer and version
        number.  The attribute has no impact on operation of the protocol,
        and serves only as a tool for diagnostic and debugging purposes.  The
        value of SOFTWARE is variable length.  It MUST be a UTF-8 [RFC3629]
        encoded sequence of less than 128 characters (which can be as long as
        763 bytes).
    """

    def __init__(self, data: bytes, description: str):
        super().__init__(data=data)
        self.__desc = description

    def __str__(self):
        return '"%s"' % self.__desc

    def __repr__(self):
        return '"%s"' % self.__desc

    @property
    def description(self) -> str:
        return self.__desc

    @classmethod
    def parse(cls, data: bytes, tag: Tag, length: Length=None):
        # check length
        if length is None or length.value == 0:
            return None
        else:
            length = length.value
            data_len = len(data)
            if data_len < length:
                # raise ValueError('data length error: %d, %d' % (data_len, length.value))
                return None
            elif data_len > length:
                data = data[:length]
        # get string
        desc = data.rstrip(b'\0').decode('utf-8')
        return cls(data=data, description=desc)

    @classmethod
    def new(cls, description: str):
        data = description.encode('utf-8')
        # padding with '\0'
        length = len(data)
        tail = length & 3
        while tail < 4:
            data += b'\0'
            tail += 1
        return cls(data=data, description=description)


#
#  Register attribute parsers
#
s_attribute_parsers = {
    MappedAddress: MappedAddressValue,
    # XorMappedAddress: XorMappedAddressValue,
    # XorMappedAddress2: XorMappedAddressValue2,

    ResponseAddress: ResponseAddressValue,
    ChangeRequest: ChangeRequestValue,
    SourceAddress: SourceAddressValue,
    ChangedAddress: ChangedAddressValue,

    Software: SoftwareValue,
}



# -*- coding: utf-8 -*-
#
#   STUN: Session Traversal Utilities for NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Session Traversal Utilities for NAT
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Client node
"""

from abc import ABC
from typing import Optional

from .protocol import Package
from .protocol import BindRequest, BindResponse
from .attributes import Attribute, AttributeLength
from .attributes import ChangePort, ChangeIPAndPort
from .attributes import ChangeRequest
from .attributes import MappedAddress, MappedAddressValue
from .attributes import XorMappedAddress, XorMappedAddressValue
from .attributes import XorMappedAddress2, XorMappedAddressValue2
from .attributes import SourceAddress, SourceAddressValue
from .attributes import ChangedAddress, ChangedAddressValue
from .attributes import Software, SoftwareValue
from .node import Node


class NatType:
    UDPBlocked = 'UDP Blocked'
    OpenInternet = 'Open Internet'
    SymmetricFirewall = 'Symmetric UDP Firewall'
    SymmetricNAT = 'Symmetric NAT'
    FullConeNAT = 'Full Cone NAT'
    RestrictedNAT = 'Restricted Cone NAT'
    PortRestrictedNAT = 'Port Restricted Cone NAT'


"""
[RFC] https://www.ietf.org/rfc/rfc3489.txt

Rosenberg, et al.           Standards Track                    [Page 21]

RFC 3489                          STUN                        March 2003


                        +--------+
                        |  Test  |
                        |   I    |
                        +--------+
                             |
                             |
                             V
                            /\              /\
                         N /  \ Y          /  \ Y             +--------+
          UDP     <-------/Resp\--------->/ IP \------------->|  Test  |
          Blocked         \ ?  /          \Same/              |   II   |
                           \  /            \? /               +--------+
                            \/              \/                    |
                                             | N                  |
                                             |                    V
                                             V                    /\
                                         +--------+  Sym.      N /  \
                                         |  Test  |  UDP    <---/Resp\
                                         |   II   |  Firewall   \ ?  /
                                         +--------+              \  /
                                             |                    \/
                                             V                     |Y
                  /\                         /\                    |
   Symmetric  N  /  \       +--------+   N  /  \                   V
      NAT  <--- / IP \<-----|  Test  |<--- /Resp\               Open
                \Same/      |   I    |     \ ?  /               Internet
                 \? /       +--------+      \  /
                  \/                         \/
                  |                           |Y
                  |                           |
                  |                           V
                  |                           Full
                  |                           Cone
                  V              /\
              +--------+        /  \ Y
              |  Test  |------>/Resp\---->Restricted
              |   III  |       \ ?  /
              +--------+        \  /
                                 \/
                                  |N
                                  |       Port
                                  +------>Restricted

                 Figure 2: Flow for type discovery process
"""


class Client(Node, ABC):

    def __init__(self, host: str, port: int):
        super().__init__(host=host, port=port)
        self.retries = 3

    def parse_attribute(self, attribute: Attribute, context: dict) -> bool:
        tag = attribute.tag
        value = attribute.value
        # check attributes
        if tag == MappedAddress:
            assert isinstance(value, MappedAddressValue), 'mapped address value error: %s' % value
            context['MAPPED-ADDRESS'] = value
        elif tag == XorMappedAddress:
            if not isinstance(value, XorMappedAddressValue):
                # XOR and parse again
                data = XorMappedAddressValue.xor(data=value.data, factor=context['trans_id'])
                length = AttributeLength(len(data))
                value = XorMappedAddressValue.parse(data=data, tag=tag, length=length)
            context['MAPPED-ADDRESS'] = value
        elif tag == XorMappedAddress2:
            if not isinstance(value, XorMappedAddressValue2):
                # XOR and parse again
                data = XorMappedAddressValue2.xor(data=value.data, factor=context['trans_id'])
                length = AttributeLength(len(data))
                value = XorMappedAddressValue2.parse(data=data, tag=tag, length=length)
            context['MAPPED-ADDRESS'] = value
        elif tag == ChangedAddress:
            assert isinstance(value, ChangedAddressValue), 'changed address value error: %s' % value
            context['CHANGED-ADDRESS'] = value
        elif tag == SourceAddress:
            assert isinstance(value, SourceAddressValue), 'source address value error: %s' % value
            context['SOURCE-ADDRESS'] = value
        elif tag == Software:
            assert isinstance(value, SoftwareValue), 'software value error: %s' % value
            context['SOFTWARE'] = value
        else:
            self.info('unknown attribute type: %s' % tag)
            return False
        self.info('%s:\t%s' % (tag, value))
        return True

    def __bind_request(self, remote_host: str, remote_port: int, body: bytes) -> Optional[dict]:
        # 1. create STUN message package
        req = Package.new(msg_type=BindRequest, body=body)
        trans_id = req.head.trans_id
        # 2. send and get response
        count = 0
        while True:
            size = self.send(data=req.data, destination=(remote_host, remote_port))
            if size != len(req.data):
                # failed to send data
                return None
            cargo = self.receive()
            if cargo is None:
                if count < self.retries:
                    count += 1
                    self.info('(%d/%d) receive nothing' % (count, self.retries))
                else:
                    # failed to receive data
                    return None
            else:
                self.info('received %d bytes from %s' % (len(cargo.data), cargo.source))
                break
        # 3. parse response
        context = {
            'trans_id': trans_id.data,
        }
        if not self.parse_data(data=cargo.data, context=context):
            return None
        head = context.get('head')
        if head is None or head.type != BindResponse or head.trans_id != trans_id:
            # received package error
            return None
        return context

    """
    [RFC] https://www.ietf.org/rfc/rfc3489.txt

    Rosenberg, et al.           Standards Track                    [Page 19]

    RFC 3489                          STUN                        March 2003


    The flow makes use of three tests.  In test I, the client sends a
    STUN Binding Request to a server, without any flags set in the
    CHANGE-REQUEST attribute, and without the RESPONSE-ADDRESS attribute.
    This causes the server to send the response back to the address and
    port that the request came from.  In test II, the client sends a
    Binding Request with both the "change IP" and "change port" flags
    from the CHANGE-REQUEST attribute set.  In test III, the client sends
    a Binding Request with only the "change port" flag set.
    """

    def __test_1(self, stun_host: str, stun_port: int) -> Optional[dict]:
        self.info('[Test 1] sending empty request ... (%s:%d)' % (stun_host, stun_port))
        body = b''
        return self.__bind_request(remote_host=stun_host, remote_port=stun_port, body=body)

    def __test_2(self, stun_host: str, stun_port: int) -> Optional[dict]:
        self.info('[Test 2] sending "ChangeIPAndPort" ... (%s:%d)' % (stun_host, stun_port))
        body = Attribute(ChangeRequest, ChangeIPAndPort).data
        return self.__bind_request(remote_host=stun_host, remote_port=stun_port, body=body)

    def __test_3(self, stun_host: str, stun_port: int) -> Optional[dict]:
        self.info('[Test 3] sending "ChangePort" ... (%s:%d)' % (stun_host, stun_port))
        body = Attribute(ChangeRequest, ChangePort).data
        return self.__bind_request(remote_host=stun_host, remote_port=stun_port, body=body)

    def get_nat_type(self, stun_host: str, stun_port: int = 3478) -> dict:
        # 1. Test I
        res1 = self.__test_1(stun_host=stun_host, stun_port=stun_port)
        if res1 is None:
            """
            The client begins by initiating test I.  If this test yields no
            response, the client knows right away that it is not capable of UDP
            connectivity.
            """
            res1 = {'NAT': NatType.UDPBlocked}
            return res1
        """
        If the test produces a response, the client examines the MAPPED-ADDRESS
        attribute.  If this address and port are the same as the local IP
        address and port of the socket used to send the request, the client
        knows that it is not NATed.  It executes test II.
        """
        ma1 = res1.get('MAPPED-ADDRESS')
        # 2. Test II
        res2 = self.__test_2(stun_host=stun_host, stun_port=stun_port)
        if ma1 is not None and (ma1.ip, ma1.port) == self.source_address:
            """
            If a response is received, the client knows that it has open access
            to the Internet (or, at least, its behind a firewall that behaves
            like a full-cone NAT, but without the translation).  If no response
            is received, the client knows its behind a symmetric UDP firewall.
            """
            if res2 is None:
                res1['NAT'] = NatType.SymmetricFirewall
                return res1
            else:
                res2['NAT'] = NatType.OpenInternet
                return res2
        elif res2 is not None:
            """
            In the event that the IP address and port of the socket did not match
            the MAPPED-ADDRESS attribute in the response to test I, the client
            knows that it is behind a NAT.  It performs test II.  If a response
            is received, the client knows that it is behind a full-cone NAT.
            """
            res2['NAT'] = NatType.FullConeNAT
            return res2
        """
        If no response is received, it performs test I again, but this time,
        does so to the address and port from the CHANGED-ADDRESS attribute
        from the response to test I.
        """
        ca1 = res1.get('CHANGED-ADDRESS')
        # 3. Test I'
        if ca1 is None:
            res1['NAT'] = 'Changed-Address not found'
            return res1
        assert isinstance(ca1, ChangedAddressValue), 'CHANGED-ADDRESS error: %s' % ca1
        changed_ip = ca1.ip
        changed_port = ca1.port
        res11 = self.__test_1(stun_host=changed_ip, stun_port=changed_port)
        if res11 is None:
            # raise AssertionError('network error')
            res1['NAT'] = 'Change address failed'
            return res1
        ma11 = res11.get('MAPPED-ADDRESS')
        if ma11 is None or ma1 is None or ma11.port != ma1.port or ma11.ip != ma1.ip:
            """
            If the IP address and port returned in the MAPPED-ADDRESS attribute
            are not the same as the ones from the first test I, the client
            knows its behind a symmetric NAT.
            """
            res11['NAT'] = NatType.SymmetricNAT
            return res11
        """
        If the address and port are the same, the client is either behind a
        restricted or port restricted NAT.  To make a determination about
        which one it is behind, the client initiates test III.  If a response
        is received, its behind a restricted NAT, and if no response is
        received, its behind a port restricted NAT.
        """
        # 4. Test III
        res3 = self.__test_3(stun_host=stun_host, stun_port=stun_port)
        if res3 is None:
            res11['NAT'] = NatType.PortRestrictedNAT
            return res11
        else:
            res3['NAT'] = NatType.RestrictedNAT
            return res3



# -*- coding: utf-8 -*-
#
#   STUN: Session Traversal Utilities for NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Session Traversal Utilities for NAT
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Server node
"""

from abc import ABC

from .protocol import Package, Header
from .protocol import BindRequest, BindResponse
from .attributes import Attribute
from .attributes import ChangePort, ChangeIPAndPort
from .attributes import ChangeRequest, ChangeRequestValue
from .attributes import MappedAddress, MappedAddressValue
from .attributes import XorMappedAddress, XorMappedAddressValue
from .attributes import XorMappedAddress2, XorMappedAddressValue2
from .attributes import SourceAddress, SourceAddressValue
from .attributes import ChangedAddress, ChangedAddressValue
from .attributes import Software, SoftwareValue
from .node import Node


class Server(Node, ABC):

    def __init__(self, host: str='0.0.0.0', port: int=3478, change_port: int=3479):
        super().__init__(host=host, port=port)
        self.software = 'stun.dim.chat 0.1'
        """
        11.2.3  CHANGED-ADDRESS

            The CHANGED-ADDRESS attribute indicates the IP address and port where
            responses would have been sent from if the "change IP" and "change
            port" flags had been set in the CHANGE-REQUEST attribute of the
            Binding Request.  The attribute is always present in a Binding
            Response, independent of the value of the flags.  Its syntax is
            identical to MAPPED-ADDRESS.
        """
        self.changed_address: (str, int) = None
        """
            "Change IP and Port"

            When this server A received ChangeRequest with "change IP" and
            "change port" flags set, it should redirect this request to the
            neighbour server B to (use another address) respond the client.

            This address will be the same as CHANGED-ADDRESS by default, but
            offer another different IP address here will be better.
        """
        self.neighbour: (str, int) = None
        """
            "Change Port"

            When this server received ChangeRequest with "change port" flag set,
            it should respond the client with another port.
        """
        self.change_port: int = change_port
        assert change_port > 0, 'change port error'
        self.hub.open(host=host, port=change_port)

    def parse_attribute(self, attribute: Attribute, context: dict) -> bool:
        tag = attribute.tag
        value = attribute.value
        # check attributes
        if tag == MappedAddress:
            assert isinstance(value, MappedAddressValue), 'mapped address value error: %s' % value
            context['MAPPED-ADDRESS'] = value
        elif tag == ChangeRequest:
            assert isinstance(value, ChangeRequestValue), 'change request value error: %s' % value
            context['CHANGE-REQUEST'] = value
        else:
            self.info('unknown attribute type: %s' % tag)
            return False
        self.info('%s: %s' % (tag, value))
        return True

    def _redirect(self, head: Header, remote_ip: str, remote_port: int) -> bool:
        """
        Redirect the request to the neighbor server

        :param remote_ip:   Client's mapped IP address
        :param remote_port: Client's mapped address port
        :return:
        """
        assert self.neighbour is not None, 'neighbour address not set'
        # create attributes
        value = MappedAddressValue.new(ip=remote_ip, port=remote_port)
        data1 = Attribute(MappedAddress, value).data
        # pack
        body = data1
        pack = Package.new(msg_type=head.type, trans_id=head.trans_id, body=body)
        res = self.send(data=pack.data, destination=self.neighbour)
        return res == pack.length

    def _respond(self, head: Header, remote_ip: str, remote_port: int, local_port: int) -> bool:
        # local (server) address
        assert self.source_address is not None, 'source address not set'
        local_ip = self.source_address[0]
        assert local_port > 0, 'local port error'
        # changed (another) address
        assert self.changed_address is not None, 'changed address not set'
        changed_ip = self.changed_address[0]
        changed_port = self.changed_address[1]
        # create attributes
        value = MappedAddressValue.new(ip=remote_ip, port=remote_port)
        data1 = Attribute(MappedAddress, value).data
        # Xor
        value = XorMappedAddressValue.new(ip=remote_ip, port=remote_port, factor=head.trans_id.data)
        data4 = Attribute(XorMappedAddress, value).data
        # Xor2
        value = XorMappedAddressValue2.new(ip=remote_ip, port=remote_port, factor=head.trans_id.data)
        data5 = Attribute(XorMappedAddress2, value).data
        # source address
        value = SourceAddressValue.new(ip=local_ip, port=local_port)
        data2 = Attribute(SourceAddress, value).data
        # changed address
        value = ChangedAddressValue.new(ip=changed_ip, port=changed_port)
        data3 = Attribute(ChangedAddress, value).data
        # software
        value = SoftwareValue.new(description=self.software)
        data6 = Attribute(Software, value).data
        # pack
        body = data1 + data2 + data3 + data4 + data5 + data6
        pack = Package.new(msg_type=BindResponse, trans_id=head.trans_id, body=body)
        res = self.send(data=pack.data, destination=(remote_ip, remote_port), source=(local_ip, local_port))
        return res == pack.length

    def handle(self, data: bytes, remote_ip: str, remote_port: int) -> bool:
        # parse request
        context = {}
        ok = self.parse_data(data=data, context=context)
        head = context.get('head')
        if not ok or head is None or head.type != BindRequest:
            # received package error
            return False
        self.info('received message type: %s' % head.type)
        change_request = context.get('CHANGE-REQUEST')
        if change_request == ChangeIPAndPort:
            # redirect for "change IP" and "change port" flags
            return self._redirect(head=head, remote_ip=remote_ip, remote_port=remote_port)
        elif change_request == ChangePort:
            # respond with another port for "change port" flag
            return self._respond(head=head, remote_ip=remote_ip, remote_port=remote_port, local_port=self.change_port)
        mapped_address = context.get('MAPPED-ADDRESS')
        if mapped_address is None:
            # respond origin request
            local_port = self.source_address[1]
            return self._respond(head=head, remote_ip=remote_ip, remote_port=remote_port, local_port=local_port)
        else:
            assert isinstance(mapped_address, MappedAddressValue), 'MAPPED-ADDRESS error: %s' % mapped_address
            # respond redirected request
            remote_ip = mapped_address.ip
            remote_port = mapped_address.port
            return self._respond(head=head, remote_ip=remote_ip, remote_port=remote_port, local_port=self.change_port)



# -*- coding: utf-8 -*-
#
#   STUN: Session Traversal Utilities for NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Session Traversal Utilities for NAT
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    [RFC] https://www.ietf.org/rfc/rfc5389.txt
"""

from udp.tlv.data import random_bytes, bytes_to_int, uint16_to_bytes, uint32_to_bytes
from udp.tlv import Data, UInt16Data


"""
                        0                 1
                        2  3  4 5 6 7 8 9 0 1 2 3 4 5

                       +--+--+-+-+-+-+-+-+-+-+-+-+-+-+
                       |M |M |M|M|M|C|M|M|M|C|M|M|M|M|
                       |11|10|9|8|7|1|6|5|4|0|3|2|1|0|
                       +--+--+-+-+-+-+-+-+-+-+-+-+-+-+

                Figure 3: Format of STUN Message Type Field
"""


class MessageType(UInt16Data):

    def __init__(self, value: int, data: bytes=None, name: str=None):
        if data is None:
            data = uint16_to_bytes(value)
        super().__init__(value=value, data=data)
        self.__name = name
        s_message_types[value] = self

    def __str__(self):
        # clazz = self.__class__.__name__
        if self.__name is None:
            return '"MessageType-0x%04X"' % self.value
        else:
            return '"%s"' % self.__name

    def __repr__(self):
        # clazz = self.__class__.__name__
        if self.__name is None:
            return '"MessageType-0x%04X"' % self.value
        else:
            return '"%s"' % self.__name

    @classmethod
    def parse(cls, data: bytes):
        data_len = len(data)
        if data_len < 2 or (data[0] & 0xC0) != 0:
            # format: 00xx xxxx, xxxx xxxx
            return None
        elif data_len > 2:
            data = data[:2]
        value = bytes_to_int(data=data)
        t = s_message_types.get(value)
        if t is None:
            return cls(value=value, data=data)
        else:
            return t


# types for a STUN message
s_message_types = {}
BindRequest = MessageType(0x0001, name='Binding Request')
BindResponse = MessageType(0x0101, name='Binding Response')
BindErrorResponse = MessageType(0x0111, name='Binding Error Response')
SharedSecretRequest = MessageType(0x0002, name='Shared Secret Request')
SharedSecretResponse = MessageType(0x0102, name='Shared Secret Response')
SharedSecretErrorResponse = MessageType(0x0112, name='Shared Secret Error Response')


class MessageLength(UInt16Data):

    def __init__(self, value: int, data: bytes=None):
        if data is None:
            data = uint16_to_bytes(value)
        super().__init__(data=data, value=value)

    @classmethod
    def parse(cls, data: bytes):
        data_len = len(data)
        if data_len < 2 or (data[1] & 0x03) != 0:
            # format: xxxx xxxx, xxxx xx00
            return None
        elif data_len > 2:
            data = data[:2]
        return super().from_bytes(data=data)


class TransactionID(Data):

    @classmethod
    def parse(cls, data: bytes):
        data_len = len(data)
        if data_len < 16:
            return None
        elif data_len > 16:
            data = data[:16]
        assert data[:4] == MagicCookie, 'transaction ID not starts with magic cookie'
        return cls(data=data)

    @classmethod
    def new(cls):
        data = MagicCookie + random_bytes(12)
        return cls(data=data)


# Magic Cookie
MagicCookie = uint32_to_bytes(0x2112A442)


"""
    STUN Message Structure
    ~~~~~~~~~~~~~~~~~~~~~~

   STUN messages are encoded in binary using network-oriented format
   (most significant byte or octet first, also commonly known as big-
   endian).  The transmission order is described in detail in Appendix B
   of RFC 791 [RFC0791].  Unless otherwise noted, numeric constants are
   in decimal (base 10).

   All STUN messages MUST start with a 20-byte header followed by zero
   or more Attributes.  The STUN header contains a STUN message type,
   magic cookie, transaction ID, and message length.
   
       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |0 0|     STUN Message Type     |         Message Length        |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Magic Cookie                          |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                                
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                            Transaction ID (96 bits)                   
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                                                                      |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

                  Figure 2: Format of STUN Message Header

   The message length MUST contain the size, in bytes, of the message
   not including the 20-byte STUN header.  Since all STUN attributes are
   padded to a multiple of 4 bytes, the last 2 bits of this field are
   always zero.  This provides another way to distinguish STUN packets
   from packets of other protocols.
"""


class Header(Data):

    def __init__(self, data: bytes, msg_type: MessageType, msg_length: MessageLength, trans_id: TransactionID):
        super().__init__(data=data)
        self.__type = msg_type
        self.__msg_length = msg_length
        self.__trans_id = trans_id

    @property
    def type(self) -> MessageType:
        return self.__type

    @property
    def msg_length(self) -> MessageLength:
        return self.__msg_length

    @property
    def trans_id(self) -> TransactionID:
        return self.__trans_id

    @classmethod
    def parse(cls, data: bytes):
        # get message type
        _type = MessageType.parse(data=data)
        if _type is None:
            return None
        pos = _type.length
        # get message length
        _len = MessageLength.parse(data=data[pos:])
        if _len is None:
            return None
        pos += _len.length
        # get transaction ID
        _id = TransactionID.parse(data=data[pos:])
        if _id is None:
            return None
        pos += _id.length
        assert pos == 20, 'header length error: %d' % pos
        if len(data) > pos:
            data = data[:pos]
        # create
        return cls(data=data, msg_type=_type, msg_length=_len, trans_id=_id)

    @classmethod
    def new(cls, msg_type: MessageType, msg_len: MessageLength, trans_id: TransactionID=None):
        if trans_id is None:
            # generate Transaction ID
            trans_id = TransactionID.new()
        # build data
        data = msg_type.data + msg_len.data + trans_id.data
        return Header(data=data, msg_type=msg_type, msg_length=msg_len, trans_id=trans_id)


class Package(Data):

    def __init__(self, data: bytes, head: Header, body: bytes):
        super().__init__(data=data)
        assert head.msg_length.value == len(body), 'STUN msg length error: %d, %d' % (head.msg_length.value, len(body))
        self.__head = head
        self.__body = body

    @property
    def head(self) -> Header:
        return self.__head

    @property
    def body(self) -> bytes:
        return self.__body

    @classmethod
    def parse(cls, data: bytes):
        # get STUN head
        head = Header.parse(data=data)
        if head is None:
            # not a STUN message?
            return None
        # check message length
        head_len = head.length
        body_len = head.msg_length.value
        pack_len = head_len + body_len
        data_len = len(data)
        if data_len < pack_len:
            # raise ValueError('STUN package length error: %d, %d' % (data_len, pack_len))
            return None
        elif data_len > pack_len:
            data = data[:pack_len]
        # get attributes body
        body = data[head_len:]
        return Package(data=data, head=head, body=body)

    @classmethod
    def new(cls, msg_type: MessageType, trans_id: TransactionID=None, body: bytes=b''):
        msg_len = MessageLength(len(body))
        head = Header.new(msg_type=msg_type, msg_len=msg_len, trans_id=trans_id)
        data = head.data + body
        return Package(data=data, head=head, body=body)



# -*- coding: utf-8 -*-
#
#   STUN: Session Traversal Utilities for NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from .protocol import *
from .attributes import *
from .node import *
from .server import *
from .client import *

name = "STUN"

__author__ = 'Albert Moky'

__all__ = [

    #
    #  Protocol
    #
    'MessageType', 'MessageLength', 'TransactionID',
    'Header', 'Package',
    # message types
    'BindRequest', 'BindResponse', 'BindErrorResponse',
    'SharedSecretRequest', 'SharedSecretResponse', 'SharedSecretErrorResponse',

    #
    #  Attributes
    #
    'Attribute', 'AttributeType', 'AttributeLength', 'AttributeValue',
    # attribute types
    'MappedAddress', 'ResponseAddress',
    'ChangeRequest',
    'SourceAddress', 'ChangedAddress',
    'Username', 'Password', 'MessageIntegrity',
    'ErrorCode', 'UnknownAttributes',
    'ReflectedFrom', 'Realm', 'Nonce',
    'XorMappedAddress', 'XorMappedAddress2', 'XorOnly',
    'Software', 'AlternateServer', 'Fingerprint',
    # attribute values
    'MappedAddressValue', 'ResponseAddressValue',
    'ChangeRequestValue', 'ChangeIP', 'ChangePort', 'ChangeIPAndPort',
    'SourceAddressValue', 'ChangedAddressValue',
    # 'UsernameValue', 'PasswordValue', 'MessageIntegrityValue',
    # 'ErrorCodeValue', 'UnknownAttributesValue',
    # 'ReflectedFromValue', 'RealmValue', 'NonceValue',
    'XorMappedAddressValue', 'XorMappedAddressValue2',  # 'XorOnlyValue',
    'SoftwareValue',  # 'AlternateServerValue', 'FingerprintValue',

    #
    #  Service
    #
    'NatType',
    'Client', 'Server',

    'get_local_ip',
]



# -*- coding: utf-8 -*-
#
#   TURN: Traversal Using Relays around NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Traversal Using Relays around NAT
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    [RFC] https://www.ietf.org/rfc/rfc5766.txt
"""

from ..attributes import AttributeType, s_attribute_parsers
from ..attributes import XorMappedAddressValue


class XorPeerAddressValue(XorMappedAddressValue):
    """
    14.3.  XOR-PEER-ADDRESS

        The XOR-PEER-ADDRESS specifies the address and port of the peer as
        seen from the TURN server.  (For example, the peer's server-reflexive
        transport address if the peer is behind a NAT.)  It is encoded in the
        same way as XOR-MAPPED-ADDRESS [RFC5389].
    """
    pass


class XorRelayedAddressValue(XorMappedAddressValue):
    """
    14.5.  XOR-RELAYED-ADDRESS

        The XOR-RELAYED-ADDRESS is present in Allocate responses.  It
        specifies the address and port that the server allocated to the
        client.  It is encoded in the same way as XOR-MAPPED-ADDRESS
        [RFC5389].
    """
    pass


# New STUN Attributes
ChannelNumber = AttributeType(0x000C, name='CHANNEL-NUMBER')
Lifetime = AttributeType(0x000D, name='LIFETIME')
# BandWidth = AttributeType(0x0010, name='BANDWIDTH')  # Reserved
XorPeerAddress = AttributeType(0x0012, name='XOR-PEER-ADDRESS')
Data = AttributeType(0x0013, name='DATA')
XorRelayedAddress = AttributeType(0x0016, name='XOR-RELAYED-ADDRESS')
EvenPort = AttributeType(0x0018, name='EVEN-PORT')
RequestedTransport = AttributeType(0x0019, name='REQUESTED-TRANSPORT')
DontFragment = AttributeType(0x001A, name='DONT-FRAGMENT')
# TimerVal = AttributeType(0x0021, name='TIMER-VAL')   # Reserved
ReservationToken = AttributeType(0x0022, name='RESERVATION-TOKEN')

#
#  Register attribute parsers
#
s_attribute_parsers[XorPeerAddress] = XorPeerAddressValue
s_attribute_parsers[XorRelayedAddress] = XorRelayedAddressValue



# -*- coding: utf-8 -*-
#
#   STUN: Session Traversal Utilities for NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Session Traversal Utilities for NAT
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Common interfaces for STUN Server or Client nodes
"""

import socket
import time
from abc import ABC, abstractmethod
from typing import Union, Optional

from udp import Hub
from udp.hub import Cargo

from .protocol import Package
from .attributes import Attribute


class Node(ABC):

    def __init__(self, port: int, host: str='0.0.0.0', hub: Hub=None):
        super().__init__()
        self.__local_address = (host, port)
        if hub is None:
            hub = Hub()
            hub.open(host=host, port=port)
            # hub.start()
        self.__hub = hub

    @property
    def source_address(self) -> tuple:
        """
        11.2.5 SOURCE-ADDRESS

            The SOURCE-ADDRESS attribute is present in Binding Responses.  It
            indicates the source IP address and port that the server is sending
            the response from.  Its syntax is identical to that of MAPPED-
            ADDRESS.


            Whether it's a server or a client, this indicates the current node's
            local address: (ip, port)
        """
        return self.__local_address

    @property
    def hub(self) -> Hub:
        return self.__hub

    def start(self):
        # start hub
        self.hub.start()

    def stop(self):
        # stop hub
        self.hub.stop()

    # noinspection PyMethodMayBeStatic
    def info(self, msg: str):
        time_array = time.localtime(int(time.time()))
        time_string = time.strftime('%y-%m-%d %H:%M:%S', time_array)
        print('[%s] %s' % (time_string, msg))

    def send(self, data: bytes, destination: tuple, source: Union[tuple, int] = None) -> int:
        """
        Send data to remote address

        :param data:
        :param destination: remote address
        :param source:      local address
        :return: count of sent bytes
        """
        try:
            return self.hub.send(data=data, destination=destination, source=source)
        except socket.error:
            return -1

    def receive(self, timeout: float=2) -> Optional[Cargo]:
        """
        Received data from local port

        :return: data and remote address
        """
        try:
            return self.hub.receive(timeout=timeout)
        except socket.error:
            return None

    @abstractmethod
    def parse_attribute(self, attribute: Attribute, context: dict) -> bool:
        """
        Parse attribute

        :param attribute:
        :param context:
        :return: False on failed
        """
        raise NotImplemented

    def parse_data(self, data: bytes, context: dict) -> bool:
        """
        Parse package data

        :param data:    data package received
        :param context: return with package head and results from body
        :return: False on failed
        """
        # 1. parse STUN package
        pack = Package.parse(data=data)
        if pack is None:
            self.info('failed to parse package data: %d' % len(data))
            return False
        # 2. parse attributes
        attributes = Attribute.parse_all(data=pack.body)
        for item in attributes:
            # 3. process attribute
            self.parse_attribute(attribute=item, context=context)
        context['head'] = pack.head
        return True


def get_local_ip(remote_host: str = '8.8.8.8', remote_port: int = 80) -> Optional[str]:
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect((remote_host, remote_port))
        ip = sock.getsockname()[0]
    finally:
        # noinspection PyUnboundLocalVariable
        sock.close()
    return ip



# -*- coding: utf-8 -*-
#
#   TURN: Traversal Using Relays around NAT
#
#                                Written in 2020 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
Mahy, et al.                 Standards Track                    [Page 5]

RFC 5766                          TURN                        April 2010


                                        Peer A
                                        Server-Reflexive    +---------+
                                        Transport Address   |         |
                                        192.0.2.150:32102   |         |
                                            |              /|         |
                          TURN              |            / ^|  Peer A |
    Client's              Server            |           /  ||         |
    Host Transport        Transport         |         //   ||         |
    Address               Address           |       //     |+---------+
   10.1.1.2:49721       192.0.2.15:3478     |+-+  //     Peer A
            |               |               ||N| /       Host Transport
            |   +-+         |               ||A|/        Address
            |   | |         |               v|T|     192.168.100.2:49582
            |   | |         |               /+-+
 +---------+|   | |         |+---------+   /              +---------+
 |         ||   |N|         ||         | //               |         |
 | TURN    |v   | |         v| TURN    |/                 |         |
 | Client  |----|A|----------| Server  |------------------|  Peer B |
 |         |    | |^         |         |^                ^|         |
 |         |    |T||         |         ||                ||         |
 +---------+    | ||         +---------+|                |+---------+
                | ||                    |                |
                | ||                    |                |
                +-+|                    |                |
                   |                    |                |
                   |                    |                |
             Client's                   |            Peer B
             Server-Reflexive    Relayed             Transport
             Transport Address   Transport Address   Address
             192.0.2.1:7000      192.0.2.15:50000     192.0.2.210:49191
"""

from .values import *

name = "TURN"

__author__ = 'Albert Moky'

__all__ = [

    #
    #  Attributes
    #

    'ChannelNumber', 'Lifetime',
    # 'BandWidth',
    'XorPeerAddress', 'Data', 'XorRelayedAddress',
    'EvenPort', 'RequestedTransport', 'DontFragment',
    # 'TimerVal',
    'ReservationToken',

    'XorPeerAddressValue', 'XorRelayedAddressValue',
]
